/**********************************************************************
 *
 * Copyright(c) 2008 Imagination Technologies Ltd. All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope it will be useful but, except 
 * as otherwise stated in writing, without any warranty; without even the 
 * implied warranty of merchantability or fitness for a particular purpose. 
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
 * 
 * The full GNU General Public License is included in this distribution in
 * the file called "COPYING".
 *
 * Contact Information:
 * Imagination Technologies Ltd. <gpl-support@imgtec.com>
 * Home Park Estate, Kings Langley, Herts, WD4 8LZ, UK 
 *
 ******************************************************************************/

#ifndef AUTOCONF_INCLUDED
#include <linux/config.h>
#endif

#include <linux/version.h>
#include <linux/module.h>
#include <linux/workqueue.h>

#include <linux/pci.h>
#include <asm/uaccess.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/interrupt.h>

#include <asm/io.h>

#include <asm/arch-omap/display.h>

#include "img_defs.h"
#include "servicesext.h"
 #include "kerneldisplay.h"
#include "omaplfb.h"

#define DRVNAME "omaplfb"

MODULE_AUTHOR("Imagination Technologies Ltd. <gpl-support@imgtec.com>");
MODULE_LICENSE("GPL");
MODULE_SUPPORTED_DEVICE(DRVNAME);



#define unref__ __attribute__ ((unused))




static int __init OMAPLFB_Init(void)
{
	if(OMAPLFBInit() != PVRSRV_OK)
		return -ENODEV;

	return 0;
}

static void __exit OMAPLFB_Cleanup(void)
{    
	if(OMAPLFBDeinit() != PVRSRV_OK)
	{
		printk ("OMAPLFB_Cleanup: can't deinit device\n");
	}
}


IMG_VOID *OMAPLFBAllocKernelMem(IMG_UINT32 ui32Size)
{
	return kmalloc(ui32Size, GFP_KERNEL);
}

IMG_VOID OMAPLFBFreeKernelMem(IMG_VOID *pvMem)
{
	kfree(pvMem);
}


PVRSRV_ERROR OMAPLFBGetLibFuncAddr (IMG_CHAR *szFunctionName, PFN_DC_GET_PVRJTABLE *ppfnFuncTable)
{
	if(strcmp("PVRGetDisplayClassJTable", szFunctionName) != 0)
		return PVRSRV_ERROR_INVALID_PARAMS;

	
	*ppfnFuncTable = PVRGetDisplayClassJTable;

	return PVRSRV_OK;
}

static void
OMAPLFBVSyncISR(void *arg, struct pt_regs *regs)
{
	OMAPLFB_SWAPCHAIN *psSwapChain= (OMAPLFB_SWAPCHAIN *)arg;
	
	(void) OMAPLFBVSyncIHandler(psSwapChain);
}

PVRSRV_ERROR OMAPLFBInstallVSyncISR(OMAPLFB_SWAPCHAIN *psSwapChain)
{

	if (omap2_disp_get_output_dev(OMAP2_GRAPHICS) == OMAP2_OUTPUT_LCD)
	 {
		if (omap2_disp_register_isr(OMAPLFBVSyncISR, psSwapChain,
					DISPC_IRQSTATUS_VSYNC))
		{
			printk("request OMAPLCD IRQ failed");
			return PVRSRV_ERROR_INIT_FAILURE;
		}
	}
	else
	if (omap2_disp_get_output_dev(OMAP2_GRAPHICS) == OMAP2_OUTPUT_TV)
	{
		if (omap2_disp_register_isr(OMAPLFBVSyncISR, psSwapChain,DISPC_IRQSTATUS_EVSYNC_EVEN|DISPC_IRQSTATUS_EVSYNC_ODD))
		{
			printk("request OMAPLCD IRQ failed");
			return PVRSRV_ERROR_INIT_FAILURE;
		}
	}
		
	return PVRSRV_OK;
}


PVRSRV_ERROR OMAPLFBUninstallVSyncISR (OMAPLFB_SWAPCHAIN *psSwapChain)
{
	omap2_disp_unregister_isr(OMAPLFBVSyncISR);
		
	return PVRSRV_OK;		
}

IMG_VOID OMAPLFBVSyncWriteReg(OMAPLFB_SWAPCHAIN *psSwapChain, IMG_UINT32 ui32Offset, IMG_UINT32 ui32Value)
{
	IMG_VOID *pvRegAddr = (IMG_VOID *)((IMG_UINT8 *)psSwapChain->pvRegs + ui32Offset);

	
	writel(ui32Value, pvRegAddr);
}

IMG_UINT32 OMAPLFBVSyncReadReg(OMAPLFB_SWAPCHAIN *psSwapChain, IMG_UINT32 ui32Offset)
{
	return readl((IMG_UINT8 *)psSwapChain->pvRegs + ui32Offset);
}

module_init(OMAPLFB_Init);
module_exit(OMAPLFB_Cleanup);

